import {CSSProperties, FC, FormEventHandler, useEffect, useRef, useState} from 'react';
import {models} from '@tonality/library-common';
import styled from 'styled-components';
import MusicalKeyboard from '@theoryofnekomata/react-musical-keyboard';
import getFormValues from '@theoryofnekomata/formxtra';
import TextInput from '../../../molecules/forms/TextInput';
import TextArea from '../../../molecules/forms/TextArea';
import ActionButton from '../../../molecules/forms/ActionButton';
import {convertToRawDuration, formatKey} from '../../../../utils/format/keyData';
import ToggleButton from '../../../molecules/forms/ToggleButton';
import NumericInput from '../../../molecules/forms/NumericInput';
import WaveOscillator from '../../../../utils/sound/WaveOscillator';
import {parseSong} from '../../../../utils/format/song';

const Form = styled('form')({
	display: 'grid',
	gap: '1rem',
});

const Toolbar = styled('div')({
	display: 'grid',
	gridTemplateColumns: '6fr 1fr 1fr 1fr',
	gap: '1rem',
	fontSize: '1.5rem',
})

const DurationSelector = styled('div')({
	display: 'grid',
	gridTemplateColumns: 'repeat(6, 1fr)',
})

const NOTE_GLYPHS = {
	1: '𝅝',
	2: '𝅗𝅥',
	4: '♩',
	8: '♪',
	16: '𝅘𝅥𝅯',
	32: '𝅘𝅥𝅰',
};

const REST_GLYPHS = {
	1: '𝄻',
	2: '𝄼',
	4: '𝄽',
	8: '𝄾',
	16: '𝄿',
	32: '𝅀',
};

type Props = {
	onSubmit?: FormEventHandler,
	action?: string,
	labels: Record<string, string>,
	defaultValues?: Partial<models.Ringtone>,
}

const CreateRingtoneForm: FC<Props> = ({
	onSubmit,
	action,
	labels,
	defaultValues = {},
}) => {
	const [hydrated, setHydrated] = useState(false);
	const [noteGlyph, setNoteGlyph] = useState(NOTE_GLYPHS[4]);
	const [restGlyph, setRestGlyph] = useState(REST_GLYPHS[4]);
	const [playTimestamp, setPlayTimestamp] = useState(0)
	const [playing, setPlaying] = useState(false)
	const [currentTime, setCurrentTime] = useState(0)
	const dataRef = useRef<HTMLTextAreaElement>(null);
	const formRef = useRef<HTMLFormElement>(null);
	const songRef = useRef({
		notes: [],
		duration: 0,
		index: 0,
	});

	const addNote = (args) => {
		if (!(args.length > 0 && dataRef.current && formRef.current)) {
			return;
		}
		const values = getFormValues(formRef.current);
		const isDotted = Boolean(values['dotted'])
		const duration = Number(values['duration']) as (1 | 2 | 4 | 8 | 16 | 32)
		const tempo = Number(values['tempo'])
		const rawDuration = convertToRawDuration(duration, tempo)
		const formattedKeyData = formatKey({
			key: args[0].key,
			dotted: isDotted,
			duration,
		});
		const newSongData = `${dataRef.current.value} ${formattedKeyData}`.trim()
		dataRef.current.value = newSongData;
		songRef.current = parseSong(newSongData, Number(values['tempo']))
		const oscillator = new WaveOscillator()
		oscillator.start(args[0].key)
		setTimeout(() => {
			oscillator.stop()
		}, rawDuration)
	};

	const addRest = () => {
		if (!(formRef.current && dataRef.current)) {
			return;
		}
		const values = getFormValues(formRef.current);
		const formattedKeyData = formatKey({
			dotted: Boolean(values['dotted']),
			duration: Number(values['duration']) as (1 | 2 | 4 | 8 | 16 | 32),
		});
		const newSongData = `${dataRef.current.value} ${formattedKeyData}`.trim()
		dataRef.current.value = newSongData;
		songRef.current = parseSong(newSongData, Number(values['tempo']))
	};

	const updateTempo = (e) => {
		if (!isNaN(e.target.valueAsNumber) && e.target.valueAsNumber > 0) {
			songRef.current = parseSong(dataRef.current.value, e.target.valueAsNumber)
		}
	}

	const togglePlayback = () => {
		if (!formRef.current) {
			return
		}
		setPlayTimestamp(Date.now())
		setPlaying(p => {
			const newValue = !p
			songRef.current.notes.forEach(n => {
				if (!isNaN(n.key)) {
					n.osc = n.osc || new WaveOscillator()
					if (!newValue && !n.stopped) {
						n.osc.stop()
					}
				}
				n.played = newValue ? false : n.played
				n.stopped = newValue ? false : n.stopped
			})
			return newValue
		})
	}

	const updateSong = () => {
		const values = getFormValues(formRef.current);
		songRef.current = parseSong(values['data'], Number(values['tempo']))
	}

	const updateView = (e) => {
		const duration = Number(e.target.value);
		setNoteGlyph(NOTE_GLYPHS[duration]);
		setRestGlyph(REST_GLYPHS[duration]);
	};

	useEffect(() => {
		let interval = setInterval(() => {
			setCurrentTime(Date.now())
		}, 50)

		return () => {
			clearInterval(interval)
		}
	}, [])

	useEffect(() => {
		if (!playing) {
			return
		}
		const cursor = currentTime - playTimestamp
		if (cursor >= songRef.current.duration) {
			setPlaying(false)
			dataRef.current.setSelectionRange(null, null)
		}

		songRef.current.notes.forEach((n) => {
			const toPlay = !n.played && n.start < cursor
			const toStop = !n.stopped && n.end < cursor
			if (toPlay) {
				if (!isNaN(n.key) && n.osc) {
					n.osc.start(n.key)
				}
				dataRef.current.select()
				dataRef.current.setSelectionRange(n.startIndex, n.endIndex)
				n.played = true
			}

			if (toStop) {
				if (!isNaN(n.key) && n.osc) {
					n.osc.stop()
				}
				n.stopped = true
			}
		})
	}, [currentTime, playTimestamp, playing])

	useEffect(() => {
		setHydrated(true);
	}, []);

	// @ts-ignore
	return (
		<Form
			onSubmit={onSubmit}
			method="post"
			action={action}
			aria-label={labels['form']}
			ref={formRef}
		>
			{
				typeof (defaultValues.id as unknown) === 'string'
				&& (
					<input
						type="hidden"
						name="id"
						defaultValue={defaultValues.id}
					/>
				)
			}
			<input
				type="hidden"
				name="composerId"
				defaultValue={defaultValues.composerId}
			/>
			<TextInput
				label={labels['name'] || 'Name'}
				name="name"
				block
				defaultValue={defaultValues.name}
			/>
			<NumericInput
				label={labels['tempo'] || 'Tempo'}
				name="tempo"
				block
				defaultValue={defaultValues.tempo || 120}
				onChange={updateTempo}
				disabled={playing}
			/>
			{
				hydrated
				&& (
					<Toolbar>
						<DurationSelector>
							<ToggleButton
								name="duration"
								value="1"
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								𝅝
							</ToggleButton>
							<ToggleButton
								name="duration"
								value="2"
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								𝅗𝅥
							</ToggleButton>
							<ToggleButton
								name="duration"
								value="4"
								defaultChecked
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								♩
							</ToggleButton>
							<ToggleButton
								name="duration"
								value="8"
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								♪
							</ToggleButton>
							<ToggleButton
								name="duration"
								value="16"
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								𝅘𝅥𝅯
							</ToggleButton>
							<ToggleButton
								name="duration"
								value="32"
								type="radio"
								onChange={updateView}
								disabled={playing}
							>
								𝅘𝅥𝅰
							</ToggleButton>
						</DurationSelector>
						<ToggleButton
							name="dotted"
							type="checkbox"
							disabled={playing}
						>
							{noteGlyph + ' + .'}
						</ToggleButton>
						<ActionButton
							onClick={addRest}
							disabled={playing}
						>
							{restGlyph}
						</ActionButton>
						<ActionButton
							onClick={togglePlayback}
						>
							{playing ? '⏹' : '▶️'}
						</ActionButton>
					</Toolbar>
				)
			}
			<TextArea
				label={labels['data'] || 'Data'}
				name="data"
				block
				defaultValue={defaultValues.data}
				readOnly={playing}
				ref={dataRef}
				onBlur={updateSong}
				style={{
					height: '12rem',
				}}
			/>
			{
				hydrated
				&& (
					<div
						style={{
							'--color-natural-key': 'var(--color-bg, white)',
							'--color-accidental-key': 'var(--color-fg, black)',
							'--color-active-key': 'Highlight',
							'--opacity-highlight': '1 !important',
						} as CSSProperties}
					>
						<MusicalKeyboard
							startKey={60}
							endKey={60 + (12 * 4) - 1}
							onChange={!playing ? addNote : undefined}
							keyLabels={k => {
								if (k % 12 === 0) {
									return (Math.floor(k / 12) - 1).toString();
								}
								return '';
							}}
							keysOn={
								playing
									? songRef.current.notes
										.filter(n => n.played && !n.stopped && !isNaN(n.key))
										.map(k => ({ key: k.key, velocity: 127, }))
									: []
							}
						/>
					</div>
				)
			}
			<ActionButton
				type="submit"
				block
			>
				{labels['cta'] || 'Post'}
			</ActionButton>
		</Form>
	);
};

export default CreateRingtoneForm;

/*
8g4 8g4 4a4 4g4 4c5 2b4 8g4 8g4 4a4 4g4 4d5 2c5 8g4 8g4 4g5 4e5 8c5 8c5 4b4 4a4 8f5 8f5 4e5 4c5 4d5 2c5.
 */
