type KeyOrRest = {
	key?: number,
	duration: 1 | 2 | 4 | 8 | 16 | 32,
	dotted?: boolean,
}

export const formatKey = ({ key, dotted, duration }: KeyOrRest) => {
	if (isNaN(key)) {
		return dotted ? `${duration}p.` : `${duration}p`
	}
	const pitchClassName = 'c,c#,d,d#,e,f,f#,g,g#,a,a#,b'.split(',')
	const pitchClass = key % 12
	const octave = Math.floor(key / 12) - 1
	return dotted ? `${duration}${pitchClassName[pitchClass]}${octave}.` : `${duration}${pitchClassName[pitchClass]}${octave}`
}

type Duration = 1 | 2 | 4 | 8 | 16 | 32
const ALLOWED_DURATIONS = [1, 2, 4, 8, 16, 32] as readonly Duration[]

type PitchClass = 'c' | 'c#' | 'd' | 'd#' | 'e' | 'f' | 'f#' | 'g' | 'g#' | 'a' | 'a#'
const PITCH_CLASSES = 'c,c#,d,d#,e,f,f#,g,g#,a,a#,b'.split(',') as readonly PitchClass[]

type Note = {
	duration: Duration,
	dotted?: boolean,
	octave?: number,
	pitchClass?: PitchClass,
}

export const parseNote = (s: string): Note => {
	const duration = parseInt(s) as Duration
	if (!ALLOWED_DURATIONS.includes(duration)) {
		throw new TypeError('Invalid duration')
	}

	const durationStr = duration.toString()
	const dotted = s.endsWith('.')
	const pitchClass = s.slice(durationStr.length, dotted ? -2 : -1) as PitchClass
	const pitchClassIndex = PITCH_CLASSES.indexOf(pitchClass)
	if (pitchClassIndex !== -1) {
		const octave = parseInt(dotted ? s.slice(-2) : s.slice(-1))
		if (!isNaN(octave)) {
			return {
				duration,
				dotted,
				octave,
				pitchClass,
			}
		}
	}
	return {
		duration,
		dotted,
	}
}

export const convertToRawDuration = (duration: number, tempo: number) => {
	if (isNaN(tempo)) {
		throw new RangeError('Tempo should not be NaN.')
	}
	if (!isFinite(tempo)) {
		throw new RangeError('Tempo should be a finite value.')
	}
	if (tempo <= 0) {
		throw new RangeError('Tempo should be a positive value.')
	}
	return (4 / duration) * (60000 / tempo)
}

export const toPlayable = (n: Note, tempo: number) => {
	const rawDurationBase = convertToRawDuration(n.duration, tempo)
	const duration = n.dotted ? rawDurationBase * 1.5 : rawDurationBase
	const pitchClassIndex = PITCH_CLASSES.indexOf(n.pitchClass)
	if (!isNaN(pitchClassIndex) && 0 <= pitchClassIndex && pitchClassIndex <= 11 && !isNaN(n.octave)) {
		return {
			duration,
			key: pitchClassIndex + (12 * (n.octave + 1))
		}
	}
	return {
		duration,
	}
}
